% A small demo for VAD presented in
% 
% [1]   Tomi Kinnunen and Padmanabhan Rajan, "A practical, self-adaptive voice 
%       activity detector for speaker verification with noisy telephone and
%       microphone data", to appear in ICASSP 2013, Vancouver.
%
% (C) Copyright School of Computing (SoC), University of Eastern Finland (UEF)
% Use this code completely at your own risk. 
%
% The code has been completely created at UEF, except for the spectral 
% subtraction code for which we have taken
% off-the-shelf implementation from Voicebox,
%
%       http://www.ee.ic.ac.uk/hp/staff/dmb/voicebox/voicebox.html
%
% So download Voicebox and add to your Matlab path.

clear all; close all;

% Load your favorite NIST speech file
[s, fs] = wavread('E:/Data/NIST-SRE-2006/data/oack.wav');

% Make sure it is mono; if you wanna apply this to 2-channel interview
% data, e.g. those in the NIST SRE 2008, 2010 and 2012, run the VAD
% separately on both channels and, for instance, use "interviewee AND NOT
% interviewer" rule to exclude false speech segments due to cross-talk from
% the interviewer. (This is certainly not optimal strategy, but here we
% focus on VAD on a given single-channel audio, not cross-talk cancellation).
s = s(:, 1);

[junk, junk, params] = VQVAD; % Get default parameters to "params" struct

% Perhaps you wanna tune some of the parameters for your custom use, such as...
% 
%   params.frame_shift      = 0.05;     % Frame hop, seconds
%   params.energy_fraction  = 0.10;     % Fraction of high/low energy frames picked for speech/nonspeech codebook training
%   params.min_energy       = -75;      % Minimum-energy constraint 
%   params.vq_size          = 16;       % Codebook size (# codewords) for speech and nonspeech

% Here we go!
tic;
[speechInd, LLR] = VQVAD(s, fs, params);

% Some stats
speech_frames    = find(speechInd);
nonspeech_frames = find(~speechInd);
n_speech         = length(speech_frames);
n_nonspeech      = length(nonspeech_frames);
speech_prop      = 100*(n_speech/(n_speech + n_nonspeech));
proc_time        = toc;
total_dur        = length(s)/fs;
rt_factor        = total_dur/proc_time;

disp(sprintf('Analyzed utterance of duration %2.2f sec, took %2.2f sec (real-time factor = %2.2f)', total_dur, proc_time, rt_factor));
disp(sprintf('Found %d speech (%2.2f %%) and %d nonspeech (%2.2f %%) frames', n_speech, speech_prop, n_nonspeech, 100 - speech_prop));

% Plot the results
subplot(311);
t = (0:length(s)-1)./fs;
plot(t, s); xlabel('Time (sec)'); ylabel('Amplitude'); axis tight;
subplot(312);
spectrogram(s, hanning(200), 100, 512, fs, 'yaxis'); xlabel('Time (sec)'); axis tight;
subplot(313);
t = (0:length(LLR)-1).*params.frame_shift;
h=plot(t,LLR); hold on;
xlabel('Time (sec)'); ylabel('LLR speech-to-nonspeech'); axis tight;